;; This Source Code Form is subject to the terms of the Mozilla Public
;; License, v. 2.0. If a copy of the MPL was not distributed with this
;; file, You can obtain one at http://mozilla.org/MPL/2.0/.
;;
;; Copyright (c) 2015-2017 Juan de la Cruz <delacruzgarciajuan@gmail.com>
;; Copyright (c) 2015-2019 Andrey Antukh <niwi@niwi.nz>

(ns uxbox.main.ui.workspace
  (:require
   [beicon.core :as rx]
   [lentes.core :as l]
   [rumext.alpha :as mf]
   [uxbox.main.constants :as c]
   [uxbox.main.data.history :as udh]
   [uxbox.main.data.undo :as udu]
   [uxbox.main.data.workspace :as dw]
   [uxbox.main.data.workspace-websocket :as dws]
   [uxbox.main.refs :as refs]
   [uxbox.main.store :as st]
   [uxbox.main.ui.confirm]
   [uxbox.main.ui.keyboard :as kbd]
   [uxbox.main.ui.messages :refer [messages-widget]]
   [uxbox.main.ui.workspace.viewport :refer [viewport]]
   [uxbox.main.ui.workspace.colorpalette :refer [colorpalette]]
   ;; [uxbox.main.ui.workspace.download]
   [uxbox.main.ui.workspace.header :refer [header]]
   ;; [uxbox.main.ui.workspace.images]
   [uxbox.main.ui.workspace.rules :refer [horizontal-rule vertical-rule]]
   [uxbox.main.ui.workspace.scroll :as scroll]
   [uxbox.main.ui.workspace.shortcuts :as shortcuts]
   [uxbox.main.ui.workspace.sidebar :refer [left-sidebar right-sidebar]]
   [uxbox.main.ui.workspace.sidebar.history :refer [history-dialog]]
   [uxbox.main.ui.workspace.streams :as uws]
   [uxbox.util.data :refer [classnames]]
   [uxbox.util.dom :as dom]
   [uxbox.util.geom.point :as gpt]
   [uxbox.util.rdnd :as rdnd]))

;; --- Workspace

(defn- on-scroll
  [event]
  (let [target (.-target event)
        top (.-scrollTop target)
        left (.-scrollLeft target)]
    (st/emit! (uws/scroll-event (gpt/point left top)))))

(defn- on-wheel
  [event canvas]
  (when (kbd/ctrl? event)
    (let [prev-zoom @refs/selected-zoom
          dom (mf/ref-node canvas)
          scroll-position (scroll/get-current-position-absolute dom)
          mouse-point @uws/mouse-position]
      (dom/prevent-default event)
      (dom/stop-propagation event)
      (if (pos? (.-deltaY event))
        (st/emit! (dw/decrease-zoom))
        (st/emit! (dw/increase-zoom)))
      (scroll/scroll-to-point dom mouse-point scroll-position))))

(mf/defc workspace-content
  [{:keys [layout page file flags] :as params}]
  (let [canvas (mf/use-ref nil)
        left-sidebar? (not (empty? (keep layout [:layers :sitemap
                                                :document-history])))
        right-sidebar? (not (empty? (keep layout [:icons :drawtools
                                                 :element-options])))
        classes (classnames
                 :no-tool-bar-right (not right-sidebar?)
                 :no-tool-bar-left (not left-sidebar?))]

     [:main.main-content
      [:section.workspace-content
       {:class classes
        :on-scroll on-scroll
        :on-wheel #(on-wheel % canvas)}

       [:& history-dialog]

       ;; Rules
       (when (contains? layout :rules)
         [:*
          [:& horizontal-rule]
          [:& vertical-rule]])

       [:section.workspace-viewport {:id "workspace-viewport" :ref canvas}
        [:& viewport {:page page}]]]

      ;; Aside
      (when left-sidebar?
        [:& left-sidebar {:file file :page page :layout layout}])
      (when right-sidebar?
        [:& right-sidebar {:page page :layout layout}])]))

(mf/defc workspace-page
  [{:keys [file-id page-id layout file flags] :as props}]
  (let [page (mf/deref refs/workspace-page)]
    [:> rdnd/provider {:backend rdnd/html5}
     [:& messages-widget]
     [:& header {:page page :layout layout :flags flags}]

     (when (:colorpalette layout)
       [:& colorpalette])

     (when (and layout page)
       [:& workspace-content {:layout layout
                              :flags flags
                              :file file
                              :page page}])]))



(mf/defc workspace
  [{:keys [file-id page-id] :as props}]
  (mf/use-effect
   {:deps (mf/deps file-id page-id)
    :fn #(st/emit! (dw/initialize file-id page-id))})

  (mf/use-effect
   {:deps (mf/deps file-id)
    :fn (fn []
          (st/emit! (dw/initialize-ws file-id))
          #(st/emit! (dw/finalize-ws file-id)))})

  ;; (mf/use-effect
  ;;  {:deps (mf/deps file-id page-id)
  ;;   :fn #(st/emit! (dw/initialize-page page-id))})

  (mf/use-effect
   {:deps (mf/deps file-id page-id)
    :fn (fn []
          (let [sub (shortcuts/init)]
            #(rx/cancel! sub)))})

  (let [layout (mf/deref refs/workspace-layout)
        flags  (mf/deref refs/selected-flags)
        file   (mf/deref refs/workspace-file)
        page   (mf/deref refs/workspace-page)]

    [:> rdnd/provider {:backend rdnd/html5}
     [:& messages-widget]
     [:& header {:page page :layout layout :flags flags}]

     (when (:colorpalette layout)
       [:& colorpalette])

     (when (and layout page)
       [:& workspace-content {:layout layout
                              :flags flags
                              :file file
                              :page page}])]))
